/*
 * Ground Control 1 resolution changer v1.0
 * Copyright 2013, 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA.
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <time.h>

typedef unsigned long int number;


// File size of gc.exe type 1
#define size_type1 2734466
// Offsets of gc.exe type 1
static number offsets_type1[ 7 ] = {
    0x03e0,
    0x03ee,
    0x03f3,
    0x2425,
    0x31e4,
    0x31f5,
    0x3833
};

// File size of gc.exe type 2
#define size_type2 2637824
// Offsets of gc.exe type 2
static number offsets_type2[ 7 ] = { 0x11d3,
    0x11e1,
    0x11e6,
    0x3275,
    0x3cd4,
    0x3ce5,
    0x4323
};

// File size of original gc.exe
#define size_typeoriginal 2680963


/*@dependent@*/ FILE *gc_exe;


#define ERR_SEEK_ERROR	"Error when seeking gc.exe\n"
#define ERR_SEEK_FERROR	"Error when seeking gc.exe\n"
#define ERR_SEEK_EOF	"Error - End of file reached when seeking gc.exe.\n"

#define ERR_READ_ERROR	"Error when reading gc.exe\n"
#define ERR_READ_FERROR	"Error when reading gc.exe\n"
#define ERR_READ_EOF	"Error - End of file reached when reading gc.exe.\n"

#define ERR_WRITE_ERROR		"Error when writing to gc.exe\n"
#define ERR_WRITE_FERROR	"Error when writing to gc.exe\n"	
#define ERR_WRITE_EOF		"Error - End of file reached when writing gc.exe.\n"



/*
================
waitforenter
================
*/
static void waitforenter()
{
    int ch;

    // Flush input
    while ((ch = getchar()) != (int)'\n' && ch != EOF);

    (void)getchar();
}

/*
================
error
Print error code and exit
================
*/
static void error(int error_code)
{
	printf("Error code: %d\n", error_code);
	waitforenter();
	exit(EXIT_FAILURE);
}

/*
================
print_byte
Prints 2 bytes pointed at by offset
================
*/
static void print_byte(number offset)
{
    unsigned char c;
    number i;

    printf("%0x: ", (unsigned int)offset);

    // Go to location pointed by offset
    if( fseek( gc_exe, (long int)offset, SEEK_SET ) != 0 ){
        printf(ERR_SEEK_ERROR);
		error(1);
    }
    if( ferror(gc_exe) != 0 ){
        printf(ERR_SEEK_FERROR);
		error(2);
    }
    if( feof(gc_exe) != 0 ){
        printf(ERR_SEEK_EOF);
		error(3);
    }

    // Print 2 bytes
    for( i = 0; i < 2; i++ ){
        if( fread( &c, 1, 1, gc_exe) == 0 ){
            printf(ERR_READ_ERROR);
			error(4);
        }
        if( ferror(gc_exe) != 0 ){
            printf(ERR_READ_FERROR);
			error(5);
        }
        if( feof(gc_exe) != 0 ){
            printf(ERR_READ_EOF);
			error(6);
        }
        printf("%.2X ", (unsigned int)c);
    }

    printf("\n");
}

/*
================
print_current
Print current bytes on screen
================
*/
static void print_current(number *offsets)
{
    print_byte(offsets[ 0 ]);
    print_byte(offsets[ 1 ]);
    print_byte(offsets[ 2 ]);
    print_byte(offsets[ 3 ]);
    print_byte(offsets[ 4 ]);
    print_byte(offsets[ 5 ]);
    print_byte(offsets[ 6 ]);
}

/*
================
write_byte
Example: If received value is 0x0640, then in binary file will be written: 40 06
Search for "little endian" to understand why
================
*/
static void write_byte(number offset, number value)
{
    unsigned char c;

    if( fseek( gc_exe, (long int)offset, SEEK_SET ) != 0 ){
        printf(ERR_SEEK_ERROR);
		error(7);
    }
    if( ferror(gc_exe) != 0 ){
        printf(ERR_SEEK_FERROR);
		error(8);
    }
    if( feof(gc_exe) != 0 ){
        printf(ERR_SEEK_EOF);
		error(9);
    }

    // Byte 1
    c = (unsigned char)value & 0xFF;
    if( fwrite( &c, 1, 1, gc_exe) == 0 ){
        printf(ERR_WRITE_ERROR);
		error(10);
    }
    if( ferror(gc_exe) != 0 ){
        printf(ERR_WRITE_FERROR);
		error(11);
    }
    if( feof(gc_exe) != 0 ){
        printf(ERR_WRITE_EOF);
		error(12);
    }

    // Byte 2
    c = (unsigned char)((int)value / 256);
    if( fwrite( &c, 1, 1, gc_exe) == 0 ){
        printf(ERR_WRITE_ERROR);
		error(13);
    }
    if( ferror(gc_exe) != 0 ){
        printf(ERR_WRITE_FERROR);
		error(14);
    }
    if( feof(gc_exe) != 0 ){
        printf(ERR_WRITE_EOF);
		error(15);
    }
}

/*
================
write_new
Set new width and height
================
*/
static void write_new(number width, number height, number *offsets)
{
    write_byte(offsets[ 0 ], width);
    write_byte(offsets[ 1 ], height);
    write_byte(offsets[ 2 ], width);
    write_byte(offsets[ 3 ], width);
    write_byte(offsets[ 4 ], width);
    write_byte(offsets[ 5 ], height);
    write_byte(offsets[ 6 ], width);
}

/*
================
filesize
================
*/
static number filesize()
{
    number size;

    if( fseek( gc_exe, 0, SEEK_END ) != 0 ){
        printf(ERR_SEEK_ERROR);
		error(16);
    }
    if( ferror(gc_exe) != 0 ){
        printf(ERR_SEEK_FERROR);
		error(17);
    }
    if( feof(gc_exe) != 0 ){
        printf(ERR_SEEK_EOF);
		error(18);
    }

    size = (number)ftell(gc_exe);

    if( fseek( gc_exe, 0, SEEK_SET ) != 0 ){
        printf(ERR_SEEK_ERROR);
		error(19);
    }
    if( ferror(gc_exe) != 0 ){
        printf(ERR_SEEK_FERROR);
		error(20);
    }
    if( feof(gc_exe) != 0 ){
        printf(ERR_SEEK_EOF);
		error(21);
    }

    return size;
}

/*
================
main
================
*/
int main()
{
    number file_size;
    number width, height;
    number *offsets;


    // Open file gc.exe for reading and writing
    gc_exe = fopen("gc.exe", "r+");
    // If there was error while opening gc.exe
    if( gc_exe == NULL ){
        printf("Error: Could not open file gc.exe\n");
		error(22);
    }
    printf("Opened file gc.exe\n");



    // File size checks
    file_size = filesize();
    printf("File size: %d\n", (int)file_size);
    if( file_size == size_type2 ){
        printf("This is cracked gc.exe type 3\n");
        offsets = offsets_type2;
    }
    else if( file_size == size_type1 ){
        printf("This is cracked gc.exe type 1\n");
        offsets = offsets_type1;
    }
    else if( file_size == size_typeoriginal ){
        printf("Error: This is the original uncracked gc.exe. Copy cracked one as gc.exe and try again.\n");
		error(23);
    }
    else{
        printf("Error: Unknown file size: %d\nCopy cracked gc.exe and try again.", (int)file_size);
		error(24);
    }



    // Print current bytes, just for fun(and debugging)
    printf("Current bytes:\n");
    print_current(offsets);



    // Get width and height values from user
    // No input validation!
    printf("Input new width: ");
    (void)scanf("%d", (int *)&width);
    if( width > 0xFFFF || width < 640 ){
        printf("Error: Invalid width\n");
		error(25);
    }
    printf("Input new height: ");
    (void)scanf("%d", (int *)&height);
    if( height > 0xFFFF || height < 480){
        printf("Error: Invalid height\n");
		error(26);
    }


    // Set width and height in gc.exe
    write_new(width, height, offsets);


    // Print new bytes, just for fun(and debugging)
    printf("New bytes:\n");
    print_current(offsets);



    // Have fun!
    printf("Resolution %dx%d successfully applied. Ground Control rocks!\n"\
"Remember to read readme.txt\n"\
"Also remember to launch the game by running launcher.bat\n"\
"Press Enter to continue...\n", (int)width, (int)height);

    waitforenter();

    return 0;
}

